from typing import Iterable


def dict_get(item, *keys):
    """
    nested dict `get()`

    >>> example={1: {2: 'X'}}
    >>> dict_get(example, 1, 2)
    'X'
    >>> dict_get(example, 1)
    {2: 'X'}
    >>> dict_get(example, 1, 2, 3) is None
    True
    >>> dict_get(example, 'foo', 'bar') is None
    True
    >>> dict_get('no dict', 'no key') is None
    True
    """
    for key in keys:
        if isinstance(item, dict):
            item = item.get(key)
        else:
            return None
    return item


def pluck(obj: dict, keys: Iterable[str]):
    """
    Extract values from a dict, if they are present

    >>> pluck({'a': 1, 'b': 2}, ['a', 'c'])
    {'a': 1}
    """
    assert isinstance(obj, dict)
    res = {}
    for k in keys:
        if k in obj:
            res[k] = obj[k]
    return res



import inspect
from pathlib import Path
from typing import Union

from bx_py_utils.path import assert_is_file


class FrameNotFound(LookupError):
    def __init__(self, file_path: Path):
        self.file_path = file_path

    def __str__(self):
        return f'Frame outside "{self.file_path}" not found!'


def last_frame_outside_path(file_path: Union[Path, str]):
    """
    Returns the stack frame that is the direct successor of given "file_path".
    The use case may be to find caller information.

    raise FrameNotFound if the given file path is not in stack.
    """
    if not isinstance(file_path, Path):
        file_path = Path(file_path)

    assert_is_file(file_path)

    frames = inspect.stack()
    if frames is None:
        # Maybe we don't run via CPython?!?
        raise RuntimeError(
            'Can not get stack frame. Current interpreter has no support for them?'
        )

    match_path = False
    for frame in frames:
        frame_path = Path(frame.filename)

        if frame_path == file_path:
            match_path = True
        elif match_path:
            return frame

    raise FrameNotFound(file_path)



class SecretsManagerMock:
    """
    Mock for `bx_py_utils.aws.secret_manager.SecretsManager()`
    """

    def __init__(self, **data):
        self.data = data

    def __call__(self, *args, **kwargs):
        return self

    def get_secret(self, secret_name, force_bytes=False):
        secret_data = self.data[secret_name]
        if force_bytes:
            secret_data = bytes(secret_data, encoding='UTF-8')
        return secret_data



import boto3


class SecretsManager:
    """
    Access AWS Secrets Manager values
    """

    def __init__(self, *, region_name):
        session = boto3.session.Session()
        self.client = session.client(
            service_name='secretsmanager',
            region_name=region_name,
        )

    def get_secret(self, secret_name, force_bytes=False, force_str=False, encoding='UTF-8'):
        response = self.client.get_secret_value(SecretId=secret_name)

        if 'SecretString' in response:
            secret_data = response['SecretString']

            if force_bytes and isinstance(secret_data, str):
                secret_data = bytes(secret_data, encoding=encoding)

        elif 'SecretBinary' in response:
            secret_data = response['SecretBinary']

            if force_str and isinstance(secret_data, bytes):
                secret_data = str(secret_data, encoding=encoding)

        else:
            raise NotImplementedError

        return secret_data



""" A simple mock for Boto3's S3 modules. """

import io
import pathlib


class PseudoBotoPaginator:
    def __init__(self, func):
        self.func = func

    def paginate(self, **kwargs):
        yield from self.func(**kwargs)


class PseudoS3Exceptions:
    class NoSuchKey(BaseException):
        def __init__(self, key):
            self.key = key

        def __repr__(self):
            return f'PseudoNoSuchKey({self.key!r})'


class PseudoBotoIO:
    # boto3's object file handle is EXTREMELY restricted. Simulate that.
    def __init__(self, content: bytes):
        assert isinstance(content, bytes)
        self._io = io.BytesIO(content)

    def read(self, *args, **kwargs):
        return self._io.read(*args, **kwargs)

    def close(self):
        return self._io.close()

    # No other properties!


class PseudoS3Client:
    """ Simulates a boto3 S3 client object in tests """

    exceptions = PseudoS3Exceptions

    def __init__(self, *, origin_client=None, init_buckets=[]):
        self.origin_client = origin_client

        self.buckets = {}
        for bucket_name in init_buckets:
            self.buckets[bucket_name] = {}

    # non-standard variable names for Boto3 compatibility
    def download_file(self, Bucket, Key, Filename):
        bucket = self.buckets[Bucket]
        if Key not in bucket:
            raise self.exceptions.NoSuchKey(Key)
        with open(Filename, 'wb') as f:
            f.write(bucket[Key])

    # non-standard variable names for Boto3 compatibility
    def download_fileobj(self, Bucket, Key, Fileobj):
        bucket = self.buckets[Bucket]
        if Key not in bucket:
            raise self.exceptions.NoSuchKey(Key)
        Fileobj.write(bucket[Key])

    # non-standard variable names for Boto3 compatibility
    def get_object(self, *, Bucket, Key):
        bucket = self.buckets[Bucket]
        if Key not in bucket:
            raise self.exceptions.NoSuchKey(Key)

        content = bucket[Key]
        return {
            'Body': PseudoBotoIO(content),
            'ContentLength': len(content),
        }

    # noqa non-standard variable names from https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.upload_fileobj
    def upload_fileobj(self, Fileobj, Bucket, Key, *, ExtraArgs=None, Callback=None, Config=None):
        if Callback:
            Callback(0)
            Callback(1)
        bucket = self.buckets[Bucket]
        contents = Fileobj.read()
        assert isinstance(bucket, dict)
        bucket[Key] = contents
        if Callback:
            Callback(len(contents))

    # noqa non-standard variable names from https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.upload_file
    def upload_file(self, Filename, Bucket, Key, *, ExtraArgs=None, Callback=None, Config=None):
        contents = pathlib.Path(Filename).read_bytes()
        buf = io.BytesIO(contents)
        self.upload_fileobj(
            Fileobj=buf, Bucket=Bucket, Key=Key, ExtraArgs=ExtraArgs, Callback=Callback,
            Config=Config)

    # noqa non-standard variable names from https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.delete_object
    def delete_object(self, Bucket, Key):
        del self.buckets[Bucket][Key]

    def generate_presigned_url(self, *args, **kwargs):
        return self.origin_client.generate_presigned_url(*args, **kwargs)

    def _list_objects_v2(self, Bucket, Prefix=None):
        bucket = self.buckets.get(Bucket, {})
        keys = bucket.keys()
        if Prefix is not None:
            keys = (k for k in keys if k.startswith(Prefix))
        keys = sorted(keys)
        yield {
            'Contents': [{'Key': k, 'Size': len(bucket[k])} for k in keys],
        }

    def get_paginator(self, operation_name):
        assert operation_name == 'list_objects_v2', f'Unsupported operation name {operation_name}'
        return PseudoBotoPaginator(self._list_objects_v2)

    # Non-standard functions, prefixed by "mock_" (can be used in test code, but not main code)
    # or "debug_" (should never be committed)
    def mock_set_content(self, Bucket, Key, content: bytes):
        assert isinstance(content, bytes)
        self.buckets[Bucket][Key] = content

    def mock_get_content(self, Bucket, Key):
        return self.buckets[Bucket][Key]

    def mock_list_files(self, Bucket):
        return sorted(self.buckets[Bucket].keys())

    def debug_long_repr(self, max_string_length=20):
        res = '<MockS3\n'
        for bucket_name, objects in sorted(self.buckets.items()):
            res += f'{bucket_name}\n'
            for key, content in sorted(objects.items()):
                try:
                    content_str = content.decode('utf-8')
                    if len(content_str) > max_string_length:
                        content_repr = repr(content_str[:max_string_length - 3])[:-1] + '...'
                    else:
                        content_repr = repr(content_str)
                except ValueError:
                    # Binary content
                    content_repr = repr(content)
                    if len(content_repr) > max_string_length:
                        content_repr = content_repr[:max_string_length - 3][:-1] + '...'

                if content == b'':
                    content_repr = ''

                res += f'  {key:20} => [{len(content_repr)} bytes] {content_repr}\n'
        res += '>\n'
        return res

    def debug_print(self):
        print(self.debug_long_repr())


class PseudoBoto3:
    def __init__(self, **kwargs):
        self.client = PseudoS3Client(**kwargs)



"""
    Assert complex output via auto updated snapshot files with nice diff error messages.
"""
import json
import pathlib
import pprint
import re
from collections import Counter
from typing import Any, Callable, Optional, Union

from bx_py_utils.html_utils import pretty_format_html, validate_html


try:
    from lxml import html  # lxml is optional requirement
except ModuleNotFoundError:
    html = None


try:
    from bs4 import BeautifulSoup  # BeautifulSoup4 is optional requirement
except ModuleNotFoundError:
    BeautifulSoup = None


from bx_py_utils.compat import removeprefix
from bx_py_utils.path import assert_is_dir
from bx_py_utils.stack_info import last_frame_outside_path
from bx_py_utils.test_utils.assertion import (
    _unified_diff,
    assert_equal,
    assert_text_equal,
    pformat_unified_diff,
    text_unified_diff,
)


SELF_FILE_PATH = pathlib.Path(__file__)
_AUTO_SNAPSHOT_NAME_COUNTER = Counter()
_UNIFY_NEWLINES_RE = r'\r?\n'


def _write_json(obj, snapshot_file):
    with snapshot_file.open('w') as snapshot_handle:
        json.dump(obj, snapshot_handle, ensure_ascii=False, indent=4, sort_keys=True)


def _get_caller_names(
        root_dir: Union[pathlib.Path, str] = None,
        snapshot_name: str = None,
        self_file_path: Union[pathlib.Path, str] = None):
    """
    Helper to get snapshot directory and name by stack frame info, but only if not given.

    self_file_path is used to get the stack frame outside of the caller for set path/filename.
    """
    if root_dir:
        if not isinstance(root_dir, pathlib.Path):
            root_dir = pathlib.Path(root_dir)
        assert_is_dir(root_dir)

    if snapshot_name is not None:
        assert re.match(r'^[-_.a-zA-Z0-9]+$', snapshot_name), (
            f'Invalid snapshot name {snapshot_name!r}'
        )

    if not root_dir or not snapshot_name:
        # Set "root_dir" and "snapshot_name" if missing by stack frame info.

        if not self_file_path:
            self_file_path = SELF_FILE_PATH

        # Get the caller stack frame (First frame that is outside this file path):
        external_stack_frame = last_frame_outside_path(file_path=self_file_path)
        caller_path = pathlib.Path(external_stack_frame.filename)
        assert caller_path.is_file()
        if not root_dir:
            root_dir = caller_path.parent

        if not snapshot_name:
            test_func_name = external_stack_frame.function
            test_func_name = removeprefix(test_func_name, 'test_')  # make it a little bit shorter

            snapshot_name = f'{caller_path.stem}_{test_func_name}'

            # Add a counter to the snapshot name, to support more than one
            # snapshot calls in the same test ;)
            counter_key = (root_dir, snapshot_name)
            _AUTO_SNAPSHOT_NAME_COUNTER[counter_key] += 1
            call_count = _AUTO_SNAPSHOT_NAME_COUNTER[counter_key]

            snapshot_name = f'{snapshot_name}_{call_count}.snapshot'

    return root_dir, snapshot_name


def _get_snapshot_file(
    root_dir: Union[pathlib.Path, str] = None,
    snapshot_name: str = None,
    extension: str = None,
    self_file_path: Union[pathlib.Path, str] = None,
):
    # Auto set dir/name via stask information, if not set:
    root_dir, snapshot_name = _get_caller_names(root_dir, snapshot_name, self_file_path)

    assert re.match(r'^[-_.a-zA-Z0-9]*$', extension), f'Invalid extension {extension!r}'

    if not snapshot_name.endswith('.snapshot'):
        snapshot_name += '.snapshot'

    snapshot_file = pathlib.Path(root_dir) / f'{snapshot_name}{extension}'
    return snapshot_file


def assert_text_snapshot(
    root_dir: Union[pathlib.Path, str] = None,
    snapshot_name: str = None,
    got: str = None,
    extension: str = '.txt',
    fromfile: str = 'got',
    tofile: str = 'expected',
    diff_func: Callable = text_unified_diff,
    self_file_path: Union[pathlib.Path, str] = None,
):
    """
    Assert "text" string via snapshot file
    """
    assert isinstance(got, str)

    snapshot_file = _get_snapshot_file(root_dir, snapshot_name, extension, self_file_path)
    try:
        expected = snapshot_file.read_bytes().decode('utf-8')
    except (FileNotFoundError, OSError):
        snapshot_file.write_bytes(got.encode('utf-8'))
        raise

    if got != expected:
        snapshot_file.write_bytes(got.encode('utf-8'))

        if re.sub(_UNIFY_NEWLINES_RE, '', got) == re.sub(_UNIFY_NEWLINES_RE, '', expected):
            raise AssertionError(f'Differing newlines: Expected {expected!r}, got {got!r}')

        # display error message with diff:
        assert_text_equal(
            got, expected,
            fromfile=fromfile, tofile=tofile,
            diff_func=diff_func
        )


def assert_snapshot(
    root_dir: Union[pathlib.Path, str] = None,
    snapshot_name: str = None,
    got: Optional[Union[dict, list]] = None,
    extension: str = '.json',
    fromfile: str = 'got',
    tofile: str = 'expected',
    diff_func: Callable = pformat_unified_diff,
    self_file_path: Union[pathlib.Path, str] = None,
):
    """
    Assert given data serialized to JSON snapshot file.
    """
    assert got is None or isinstance(got, (dict, list)), \
        f'Not JSON-serializable: {got!r} is not a dict or list, but a {type(got).__name__}'

    snapshot_file = _get_snapshot_file(root_dir, snapshot_name, extension, self_file_path)
    try:
        with snapshot_file.open('r') as snapshot_handle:
            expected = json.load(snapshot_handle)
    except (ValueError, OSError, FileNotFoundError):
        _write_json(got, snapshot_file)
        raise

    if got != expected:
        _write_json(got, snapshot_file)

        # display error message with diff:
        assert_equal(
            got, expected,
            fromfile=fromfile, tofile=tofile,
            diff_func=diff_func
        )


def assert_py_snapshot(
    root_dir: Union[pathlib.Path, str] = None,
    snapshot_name: str = None,
    got: Any = None,
    extension: str = '.txt',
    fromfile: str = 'got',
    tofile: str = 'expected',
    diff_func: Callable = _unified_diff,
    self_file_path: Union[pathlib.Path, str] = None,
):
    """
    Assert complex python objects vio PrettyPrinter() snapshot file.

    Advantage over JSON:
     - More python object types are supported
     - The comparison is stricter. e.g.: UUID object instance vs. UUID string
    """
    got_str = pprint.pformat(got, indent=4, width=120)

    snapshot_file = _get_snapshot_file(root_dir, snapshot_name, extension, self_file_path)
    try:
        expected_str = snapshot_file.read_text()
    except FileNotFoundError:
        snapshot_file.write_text(got_str)
        raise

    if got_str != expected_str:
        snapshot_file.write_text(got_str)

        # display error message with diff:
        assert_equal(
            got_str, expected_str,
            fromfile=fromfile, tofile=tofile,
            diff_func=diff_func
        )


def assert_html_snapshot(
    root_dir: Union[pathlib.Path, str] = None,
    snapshot_name: str = None,
    got: str = None,
    extension: str = '.html',
    fromfile: str = 'got',
    tofile: str = 'expected',
    diff_func: Callable = text_unified_diff,
    self_file_path: Union[pathlib.Path, str] = None,
    validate: bool = True,
    validate_kwargs: dict = None,
    pretty_format: bool = True,
    pretty_kwargs: dict = None,
):
    """
    Assert "html" string via snapshot file with validate and pretty format
    """
    assert isinstance(got, str)

    if validate:
        if validate_kwargs is None:
            validate_kwargs = {}
        validate_html(got, **validate_kwargs)

    if pretty_format:
        if pretty_kwargs is None:
            pretty_kwargs = {}
        got = pretty_format_html(got, **pretty_kwargs)

    assert_text_snapshot(
        root_dir=root_dir,
        snapshot_name=snapshot_name,
        got=got,
        extension=extension,
        fromfile=fromfile,
        tofile=tofile,
        diff_func=diff_func,
        self_file_path=self_file_path,
    )



import difflib

from bx_py_utils.humanize.pformat import pformat


def text_ndiff(txt1, txt2, fromfile=None, tofile=None):
    """
    Generate a `ndiff` between two text strings.
    (fromfile/tofile are ignored, only for compatibility with text_unified_diff signature)
    """
    return '\n'.join(difflib.ndiff(txt1.splitlines(), txt2.splitlines()))


def pformat_ndiff(obj1, obj2, fromfile=None, tofile=None):
    """
    Generate a `ndiff` from two objects, using `pformat()`
    (fromfile/tofile are ignored, only for compatibility with pformat_unified_diff signature)
    """
    return '\n'.join(
        difflib.ndiff(
            pformat(obj1).splitlines(),
            pformat(obj2).splitlines()
        )
    )


def _unified_diff(txt1, txt2, fromfile: str = 'got', tofile: str = 'expected'):
    assert isinstance(txt1, str)
    assert isinstance(txt2, str)
    return '\n'.join(
        difflib.unified_diff(
            txt1.splitlines(), txt2.splitlines(),
            fromfile=fromfile, tofile=tofile
        )
    )


def text_unified_diff(txt1, txt2, fromfile: str = 'got', tofile: str = 'expected'):
    """
    Generate a unified diff between two text strings.
    """
    return _unified_diff(txt1, txt2, fromfile=fromfile, tofile=tofile)


def pformat_unified_diff(obj1, obj2, fromfile: str = 'got', tofile: str = 'expected'):
    """
    Generate a unified diff from two objects, using `pformat()`
    """
    return _unified_diff(pformat(obj1), pformat(obj2), fromfile=fromfile, tofile=tofile)


def assert_equal(
        obj1,
        obj2,
        msg=None,
        fromfile: str = 'got',
        tofile: str = 'expected',
        diff_func=pformat_unified_diff):
    """
    Check if the two objects are the same. Display a nice diff, using `pformat()`
    """
    if obj1 != obj2:
        if not msg:
            msg = 'Objects are not equal:'
        raise AssertionError(f'{msg}\n{diff_func(obj1,obj2, fromfile=fromfile, tofile=tofile)}')


def assert_text_equal(
        txt1,
        txt2,
        msg=None,
        fromfile: str = 'got',
        tofile: str = 'expected',
        diff_func=text_unified_diff):
    """
    Check if the two text strings are the same. Display a error message with a diff.
    """
    assert isinstance(txt1, str)
    assert isinstance(txt2, str)
    if txt1 != txt2:
        if not msg:
            msg = 'Text not equal:'
        raise AssertionError(f'{msg}\n{diff_func(txt1, txt2, fromfile=fromfile, tofile=tofile)}')



import hashlib
import string


# Remove umlauts will hopefully avoid to generate valid words ;)
ALPHABET = (
    '-._~'  # Allowed additional characters RFC 3986
    'bcdfghjklmnpqrstvwxyz'  # lower case letters without aeiou
    'BCDFGHJKLMNPQRSTVWXYZ'  # upper case letters without AEIOU
) + string.digits


def url_safe_encode(data, alphabet=ALPHABET):
    """
    Encode bytes into a URL safe string.
    Note:
        Use a URL safe alphabet (see RFC 3986) without umlauts
    """
    assert isinstance(data, bytes)

    len_alphabet = len(alphabet)
    return ''.join(alphabet[char % len_alphabet] for char in data)


def url_safe_hash(data, max_size=None, hasher_name='sha3_512', encoding='utf-8'):
    """
    Generate a URL safe hash with `max_size` from given string/bytes.

    >>> url_safe_hash('foo', max_size=16)
    'tMXtn6KpcjzTdzTk'
    """
    if isinstance(data, str):
        data = bytes(data, encoding=encoding)

    # Generate hash digest:
    hasher = hashlib.new(hasher_name)
    hasher.update(data)
    hash_digest = hasher.digest()

    # Convert hash digest bytes into URL safe string:
    safe_hash = url_safe_encode(hash_digest)
    if max_size:
        assert len(safe_hash) >= max_size, 'Hash digest too short for requested max size!'
        safe_hash = safe_hash[:max_size]

    return safe_hash



from json import JSONDecodeError
from pathlib import Path

from bx_py_utils.test_utils.assertion import assert_equal
from bx_py_utils.test_utils.snapshot import assert_snapshot


def build_requests_mock_history(mock, only_json=True):
    history = []
    for request in mock.request_history:

        request_info = {
            'request': f'{request.method} {request.url}'
        }
        try:
            json_data = request.json()
        except JSONDecodeError as err:
            if only_json:
                raise AssertionError(
                    f'{request.method} {request.url} without valid JSON: {err} in:\n{err.doc!r}'
                )
            request_info['text'] = request.text
        else:
            request_info['json'] = json_data

        history.append(request_info)
    return history


def assert_json_requests_mock(mock, data):
    """
    Check the requests mock history. In this case all requests must be JSON.
    e.g.:

        with requests_mock.mock() as m:
            m.post('http://test.tld', text='resp')
            requests.post('http://test.tld', json={'foo': 'bar'})

        assert_json_requests_mock(mock=m, data=[{
            'request': 'POST http://test.tld/',
            'json': {'foo': 'bar'},
        }])
    """
    history = build_requests_mock_history(mock, only_json=True)
    assert_equal(history, data, msg='Request history are not equal:')


def assert_json_requests_mock_snapshot(mock):
    """
    Check requests mock history via snapshot. Accepts only JSON requests.
    :param mock:
    :return:
    """
    history = build_requests_mock_history(mock, only_json=True)
    assert_snapshot(got=history, self_file_path=Path(__file__))


def assert_requests_mock(mock, data):
    """
    Check the requests mock history. Accept mixed "text" and "JSON".
    e.g.:

        with requests_mock.mock() as m:
            m.get('http://test.tld', text='foo')
            m.post('http://test.tld', text='bar')
            requests.post('http://test.tld', data={'foo': 'one'})
            requests.post('http://test.tld', json={'foo': 'two'})

        assert_requests_mock(mock=m, data=[{
            'request': 'POST http://test.tld/',
            'text': 'foo=one',
        }, {
            'request': 'POST http://test.tld/',
            'json': {'foo': 'two'},
        }])
    """
    history = build_requests_mock_history(mock, only_json=False)
    assert_equal(history, data, msg='Request history are not equal:')


def assert_requests_mock_snapshot(mock):
    """
    Check requests mock history via snapshot. Accept mixed "text" and "JSON".
    """
    history = build_requests_mock_history(mock, only_json=False)
    assert_snapshot(got=history, self_file_path=Path(__file__))



import sys


TRACEBACK_MAX_CHARS = 100

DEFAULT_STOP_ON_FILE_PATH = (
    '/django/core/handlers/base.py',
    'django/core/management/base.py',
    'django/core/handlers/exception.py',
    f'/python{".".join(str(v) for v in sys.version_info[:2])}/wsgiref'
)


def print_exc_plus(exc=None, stop_on_file_path=None, max_chars=None):
    """ Print traceback information with a listing of all the local variables in each frame. """
    if exc is None:
        tb = sys.exc_info()[2]
    else:
        tb = exc.__traceback__  # FIXME: How to get the Traceback object better?

    while tb.tb_next:
        tb = tb.tb_next
    stack = []
    f = tb.tb_frame
    while f:
        stack.append(f)
        f = f.f_back

    print_local_vars = True

    if stop_on_file_path is None:
        stop_on_file_path = DEFAULT_STOP_ON_FILE_PATH

    if max_chars is None:
        max_chars = TRACEBACK_MAX_CHARS

    print(' -' * 50, file=sys.stderr)
    print('Locals by frame, most recent call first:', file=sys.stderr)
    for frame in stack:
        file_path = frame.f_code.co_filename
        print(
            f'\n File "{file_path}", line {frame.f_lineno}, in {frame.f_code.co_name}',
            end='',
            flush=True,
            file=sys.stderr)

        if stop_on_file_path and print_local_vars:
            for path_part in stop_on_file_path:
                if path_part in file_path:
                    print_local_vars = False
                    break

        if print_local_vars:
            print(file=sys.stderr)
            for key, value in list(frame.f_locals.items()):
                print('%30s =' % key, end=' ', file=sys.stderr)

                # We have to be careful not to cause a new error in our error
                # printer! Calling str() on an unknown object could cause an
                # error we don't want.
                value = repr(value)

                if len(value) + 3 > max_chars:
                    value = f'{value[:max_chars-3]}...'

                try:
                    print(value, file=sys.stderr)
                except BaseException:
                    print('<ERROR WHILE PRINTING VALUE>', file=sys.stderr)

    print(file=sys.stderr)
    print('=' * 100, file=sys.stderr)



import datetime


def parse_dt(dtstr):
    """
    Helper for easy generate a `datetime` instance via string.

    >>> parse_dt(None) is None
    True
    >>> parse_dt('2020-01-01T00:00:00+0000')
    datetime.datetime(2020, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    if dtstr is None:
        return None
    return datetime.datetime.strptime(dtstr, '%Y-%m-%dT%H:%M:%S%z')



import inspect
import json


def introspection_query(depth: int):
    """
    Generate GraphQL introspection query with variable nested depth.
    """
    fields_query = 'kind name ofType {' * depth + 'kind name' + '}' * depth
    return inspect.cleandoc(
        '''{
            __schema {
                types {
                    kind
                    name
                    fields {
                        name
                        type {
                            ''' + fields_query + '''
                        }
                    }
                }
            }
        }'''
    )


def _resolve_type(types_by_name, graphql_type):
    crumbs = []

    while True:
        crumbs.append(graphql_type)
        if len(crumbs) >= 99:
            crumbs_repr = '\n'.join(
                f'{i}. {json.dumps(crumb)}' for i, crumb in enumerate(crumbs, start=1))

            assert len(crumbs) < 99, f'Aborting actualization (closed loop?)\nStack:\n{crumbs_repr}'

        if graphql_type['kind'] in ('LIST', 'NON_NULL'):
            graphql_type = graphql_type['ofType']
            continue

        if 'ofType' in graphql_type and graphql_type['ofType'] is None and graphql_type.get('name'):
            # Named reference to some other type
            referenced_name = graphql_type['name']
            referenced_type = types_by_name.get(referenced_name)
            assert referenced_type, f'Cannot resolve type name {referenced_name}'
            graphql_type = referenced_type
            continue

        if graphql_type['kind'] == 'UNION':
            return graphql_type

        return graphql_type


def _craft_complete_query(types_by_name, visited_types, graphql_type, allow_recursive,
                          max_depth=None, indent=''):
    graphql_type = _resolve_type(types_by_name, graphql_type)

    if graphql_type in visited_types:
        return '# Skipping already visited type'

    kind = graphql_type['kind']
    if kind == 'SCALAR':
        return ''
    elif kind == 'ENUM':
        return '# Skipping enum'
    elif kind == 'UNION':
        return '# Skipping union'
    elif kind == 'INTERFACE':
        return '# Skipping interface'
    elif kind == 'OBJECT':
        if max_depth is not None and max_depth <= 0:
            return '# More details hidden by max_depth'
        next_max_depth = max_depth if max_depth is None else max_depth - 1
        visited_types.append(graphql_type)
        fields = sorted(graphql_type['fields'], key=lambda t: t['name'])
        return '{\n' + '\n'.join(
            (f'{indent}  {field["name"]} ' + _craft_complete_query(
                types_by_name, visited_types[:] if allow_recursive else visited_types,
                field['type'], allow_recursive, next_max_depth, indent + '  '
            ))
            for field in fields
        ) + '\n' + indent + '}'
    else:
        raise ValueError(f'Unsupported GraphQL kind {kind}')


def complete_query(schema_doc: dict, root_name: str, allow_recursive=True, max_depth=None) -> str:
    types_by_name = {}
    for graphql_type in schema_doc['__schema']['types']:
        types_by_name[graphql_type['name']] = graphql_type

    return _craft_complete_query(
        types_by_name, [], types_by_name[root_name],
        allow_recursive=allow_recursive, max_depth=max_depth)



from bx_py_utils.text_tools import cutout


try:
    from lxml import etree, html  # lxml is optional requirement
except ModuleNotFoundError:
    html = None
else:
    from lxml.etree import XMLSyntaxError


try:
    from bs4 import BeautifulSoup  # BeautifulSoup4 is optional requirement
except ModuleNotFoundError:
    BeautifulSoup = None


class InvalidHtml(AssertionError):
    """
    XMLSyntaxError with better error messages: used in validate_html()
    """

    def __init__(self, *args):
        self.args = args

        data, origin_err = args
        assert isinstance(data, str)
        assert isinstance(origin_err, XMLSyntaxError)

        self.origin_msg = origin_err.msg

        line_no, column = origin_err.position
        self.cutout_text = cutout(data, line_no, column, extra_lines=3)

    def __str__(self):
        return (
            f'{self.origin_msg}\n'
            f'{"-"*80}\n'
            f'{self.cutout_text}\n'
            f'{"-"*80}'
        )


def validate_html(data, **parser_kwargs):
    """
    Validate a HTML document via XMLParser (Needs 'lxml' package)

    There are a few more ways to validate HTML documents,
    but the intention here is just to raise an error on
    really broken documents.

    To be more strict set "recover=False"
    """
    assert isinstance(data, str)

    if html is None:
        raise ModuleNotFoundError(
            'This feature needs "lxml", please add it to you requirements'
        )

    parser = etree.XMLParser(**parser_kwargs)
    try:
        parser.feed(data)
        parser.close()
    except XMLSyntaxError as err:
        raise InvalidHtml(data, err)


def pretty_format_html(data, parser='html.parser', **bs_kwargs):
    """
    Pretty format given HTML document via BeautifulSoup (Needs 'beautifulsoup4' package)
    """
    assert isinstance(data, str)

    if BeautifulSoup is None:
        raise ModuleNotFoundError(
            'This feature needs "beautifulsoup4", please add it to you requirements'
        )

    soup = BeautifulSoup(data, parser, **bs_kwargs)
    return soup.prettify(
        formatter=None  # Do not perform any substitution
    )



def removeprefix(self: str, prefix: str) -> str:
    """ Backport of `removeprefix` from PEP-616 (Python 3.9+) """

    if self.startswith(prefix):
        return self[len(prefix):]
    else:
        return self


def removesuffix(self: str, suffix: str) -> str:
    """ Backport of `removesuffix` from PEP-616 (Python 3.9+) """

    if self.endswith(suffix):
        return self[:-len(suffix)]
    else:
        return self



import importlib
import inspect
import re
import sys
import warnings
from pathlib import Path


try:
    from pdoc import extract
    from pdoc.doc import Module
except ImportError:
    if sys.version_info >= (3, 7):
        # pdoc is not compatible with Python 3.6
        raise


from bx_py_utils.test_utils.assertion import assert_text_equal


class ModulePath:
    def __init__(self):
        self.cache = {}

    def get_base_path(self, module_name):
        base_module = module_name.partition('.')[0]
        if base_module not in self.cache:
            module = importlib.import_module(base_module)
            module_path = Path(module.__file__).parent
            self.cache[base_module] = module_path
            return module_path
        return self.cache[base_module]


def get_code_location(obj):
    """
    Return start and end line number for an object via inspect.
    """
    lines, start = inspect.getsourcelines(obj)
    return start, start + len(lines) - 1


def generate_modules_doc(modules, start_level=1, link_template=None):
    """
    Generate a list of function/class information via pdoc.
    """
    assert sys.version_info >= (3, 7), 'pdoc is not compatible with Python 3.6'

    if link_template and 'lnum' in link_template:
        warnings.warn(
            '"lnum" in "link_template" will be removed in the future. Change it to "start"',
            DeprecationWarning
        )

    def first_doc_line(doc_string):
        try:
            return doc_string.splitlines()[0]
        except IndexError:
            return None

    def get_doc_list(pdoc_list, root_path):
        for pdoc_item in sorted(pdoc_list):
            item_name = pdoc_item.name
            if item_name.startswith('_'):
                continue

            item_doc_line = first_doc_line(pdoc_item.docstring)
            if not item_doc_line:
                continue

            item_name = f'`{item_name}()`'

            if link_template:
                item_path = Path(pdoc_item.source_file)
                path = item_path.relative_to(root_path)

                start, end = get_code_location(obj=pdoc_item.obj)
                link = link_template.format(
                    path=path,
                    lnum=start,  # TODO: Obsolete, will be removed in the future!
                    start=start,
                    end=end,
                )
                item_name = f'[{item_name}]({link})'

            yield f'* {item_name} - {item_doc_line}\n'

    module_path = ModulePath()

    module_names = extract.parse_specs(modules=modules)
    parts = []
    for module_name in sorted(module_names):
        module_obj = extract.load_module(module_name)

        base_path = module_path.get_base_path(module_name=module_name)
        module_root_path = base_path.parent

        pdoc_module = Module(module_obj)

        # Collect information

        module_doc_line = first_doc_line(pdoc_module.docstring)

        class_docs = list(get_doc_list(pdoc_list=pdoc_module.classes, root_path=module_root_path))
        func_docs = list(get_doc_list(pdoc_list=pdoc_module.functions, root_path=module_root_path))

        # Generate output only if information exists:

        if module_doc_line or class_docs or func_docs:
            level = module_name.count('.') + start_level
            parts.append(f'\n{"#" * level} {module_name}\n\n')

        if module_doc_line:
            parts.append(f'{module_doc_line}\n\n')

        parts.extend(class_docs)
        parts.extend(func_docs)
    return ''.join(parts)


def assert_readme(
    readme_path: Path,
    modules: list,
    start_marker_line: str = '[comment]: <> (✂✂✂ auto generated start ✂✂✂)',
    end_marker_line: str = '[comment]: <> (✂✂✂ auto generated end ✂✂✂)',
    start_level: int = 1,
    link_template: str = None,
):
    """
    Check and update README file with generate_modules_doc()
    The automatic generated documentation list will be "replace" between the given markers.
    """
    assert sys.version_info >= (3, 7), 'pdoc is not compatible with Python 3.6'

    assert readme_path.is_file()
    old_readme = readme_path.read_text()

    assert start_marker_line in old_readme
    assert end_marker_line in old_readme

    doc_block = generate_modules_doc(
        modules=modules,
        start_level=start_level,
        link_template=link_template,
    )

    doc_block = f'{start_marker_line}\n{doc_block}\n{end_marker_line}'

    start = re.escape(start_marker_line)
    end = re.escape(end_marker_line)

    new_readme, sub_count = re.subn(
        f'{start}(.*?){end}',
        doc_block,
        old_readme,
        flags=re.DOTALL
    )
    assert sub_count == 1
    if old_readme != new_readme:
        readme_path.write_text(new_readme)

        # display error message with diff:
        assert_text_equal(old_readme, new_readme)



import collections


class MockedBoto3Session:
    """
    Mock for `boto3.session.Session()`
    but only this parts:
     * session.client()
     * session.client().get_secret_value()
    """

    def __init__(self):
        self._secrets = collections.defaultdict(dict)

        self.client = self  # fake this call: session.client()

    def __call__(self, *args, **kwargs):
        return self

    ##########################################################################
    # methods to load mock data:

    def add_secret_string(self, **kwargs):
        assert kwargs
        for key, value in kwargs.items():
            assert isinstance(key, str)
            assert isinstance(value, str)
            self._secrets[key]['SecretString'] = value

    def add_secret_binary(self, **kwargs):
        assert kwargs
        for key, value in kwargs.items():
            assert isinstance(key, str)
            assert isinstance(value, bytes)
            self._secrets[key]['SecretBinary'] = value

    ##########################################################################
    # replicated API from boto3.session.Session()

    def get_secret_value(self, SecretId):
        try:
            secret_value_response = self._secrets[SecretId]
        except KeyError:
            raise KeyError(f'Mock value for secret {SecretId!r} missing!')

        return secret_value_response



import tempfile

from bx_py_utils.aws.secret_manager import SecretsManager


class ClientSideCertManager:
    """
    Helper to manage client-side TLS certificate via AWS Secrets Manager by
    storing the cert/key content in a temporary file.

    Libraries like "requests" any many others needs the client-side TLS certificate as files.
    They don't accept cert/key as strings. So we have to fetch the content from AWS and
    store them in a file.

    Usage e.g.:

        with ClientSideCertManager(
            tls_cert_secret_name='foo',
            tls_key_secret_name='bar',
            region_name='eu-central-1'
        ) as cert_manager:
            response = requests.get(
                'https://foo.tld/bar/',
                cert=cert_manager.cert,
            )
    """

    def __init__(self, *, tls_cert_secret_name, tls_key_secret_name, region_name):
        self._tls_cert_secret_name = tls_cert_secret_name
        self._tls_key_secret_name = tls_key_secret_name

        self._secrets_manager = SecretsManager(region_name=region_name)
        self._tls_cert_file = None
        self._tls_key_file = None

        self.cert = None

    def _store_secret(self, secret_name, temp_file):
        secret = self._secrets_manager.get_secret(
            secret_name=secret_name,
            force_bytes=True
        )
        assert secret, f'No {secret_name!r} secret !'

        temp_file.write(secret)
        temp_file.flush()

    def __enter__(self):
        # Store TLS certificate:
        self._tls_cert_file = tempfile.NamedTemporaryFile(prefix='tls_cert_')
        self._store_secret(
            secret_name=self._tls_cert_secret_name,
            temp_file=self._tls_cert_file,
        )

        # Store TLS key:
        self._tls_key_file = tempfile.NamedTemporaryFile(prefix='tls_key_')
        self._store_secret(
            secret_name=self._tls_key_secret_name,
            temp_file=self._tls_key_file,
        )

        self.cert = (self._tls_cert_file.name, self._tls_key_file.name)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._tls_cert_file is not None:
            self._tls_cert_file.close()

        if self._tls_key_file is not None:
            self._tls_key_file.close()



# https://gist.github.com/Chiron1991/8199fc1a41c2107982053aba809838c6
#
# tests functions from the gist were moved to utilities.tests.test_processify
# so they can be picked up by our test runner

import sys
import traceback
from functools import wraps
from multiprocessing import Process, Queue


def processify(func):
    """
    Decorator to run a function as a process.
    Be sure that every argument and the return value
    is *pickable*.
    The created process is joined, so the code does not
    run in parallel.
    """

    def process_func(q, *args, **kwargs):
        try:
            ret = func(*args, **kwargs)
        except Exception:
            ex_type, ex_value, tb = sys.exc_info()
            error = ex_type, ex_value, ''.join(traceback.format_tb(tb))
            ret = None
        else:
            error = None

        q.put((ret, error))

    # register original function with different name
    # in sys.modules so it is pickable
    process_func.__name__ = func.__name__ + 'processify_func'
    setattr(sys.modules[__name__], process_func.__name__, process_func)

    @wraps(func)
    def wrapper(*args, **kwargs):
        q = Queue()
        p = Process(target=process_func, args=(q,) + args, kwargs=kwargs)
        p.start()
        ret, error = q.get()
        p.join()

        if error:
            ex_type, ex_value, tb_str = error
            message = f'{str(ex_value)} (in subprocess)\n{tb_str}'
            raise ex_type(message)

        return ret

    return wrapper



from pathlib import Path

from bx_py_utils.path import assert_is_dir


class FileWatcher:
    """
    Helper to record which new files have been created.
    Currently the focus is:
     * Watch only one directory
     * Cleanup/remove only files
    """

    def __init__(self, base_path: Path, cleanup: bool = True):
        assert isinstance(base_path, Path)
        assert_is_dir(base_path)

        self.base_path = base_path
        self.cleanup = cleanup

        self.initial_items = None

    def _get_items(self):
        items = set(self.base_path.glob('*'))
        return items

    def __enter__(self):
        self.initial_items = self._get_items()
        return self

    def get_new_items(self):
        current_items = self._get_items()
        new_items = current_items - self.initial_items
        return new_items

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.cleanup:
            # Cleanup -> remove all created files
            new_items = self.get_new_items()
            errors = []
            for item in new_items:
                if item.exists():
                    try:
                        item.unlink()
                    except Exception as err:
                        errors.append(str(err))

            if errors:
                raise RuntimeError(', '.join(errors))



import itertools


def chunk_iterable(iterable, chunk_size):
    """
    Returns a generator that yields slices of iterable of the given `chunk_size`.

    Basic example:
    >>> list(chunk_iterable(range(6), 2))
    [(0, 1), (2, 3), (4, 5)]

    Remaining elements that don't fill a full chunk are returned gracefully:
    >>> list(chunk_iterable(range(5), 2))
    [(0, 1), (2, 3), (4,)]
    """
    iterator = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(iterator, chunk_size))
        if not chunk:
            return  # exit the endless loop!
        yield chunk



import datetime


TIMESINCE_CHUNKS = (
    (60 * 60 * 24 * 365, '%.1f\xa0years'),
    (60 * 60 * 24 * 30, '%.1f\xa0months'),
    (60 * 60 * 24 * 7, '%.1f\xa0weeks'),
    (60 * 60 * 24, '%.1f\xa0days'),
    (60 * 60, '%.1f\xa0hours'),
    (60, '%.1f\xa0minutes'),
)


def human_timedelta(t):
    """
    Converts a time duration into a friendly text representation.

    >>> human_timedelta(0.001)
    '1.0\xa0ms'
    >>> human_timedelta(0.01)
    '10.0\xa0ms'
    >>> human_timedelta(0.9)
    '900.0\xa0ms'
    >>> human_timedelta(1)
    '1.0\xa0seconds'
    >>> human_timedelta(65.5)
    '1.1\xa0minutes'
    >>> human_timedelta(59 * 60)
    '59.0\xa0minutes'
    >>> human_timedelta(1.05*60*60)
    '1.1\xa0hours'
    >>> human_timedelta(24*60*60)
    '1.0\xa0days'
    >>> human_timedelta(2.54 * 60 * 60 * 24 * 365)
    '2.5\xa0years'
    """
    if isinstance(t, datetime.timedelta):
        t = t.total_seconds()
    assert isinstance(t, (int, float))

    if abs(t) < 1:
        return f'{round(t * 1000, 1):.1f}\xa0ms'
    if abs(t) < 60:
        return f'{round(t, 1):.1f}\xa0seconds'

    for seconds, time_string in TIMESINCE_CHUNKS:
        count = t / seconds
        if abs(count) >= 1:
            count = round(count, 1)
            break

    return time_string % count



def cutout(text, line_no, column, extra_lines=2):
    """
    Mark a point in a long text by line no + column with context lines around.
    """
    assert isinstance(text, str)
    assert line_no >= 0
    assert column >= 0
    assert extra_lines >= 0

    lines = text.splitlines()
    line_count = len(lines)

    assert line_no <= line_count

    from_line = line_no - extra_lines - 1
    if from_line < 0:
        from_line = 0

    to_line = line_no + extra_lines
    if to_line > line_count:
        to_line = line_count

    line_no_width = len(str(from_line)) + 1

    lines = lines[from_line: to_line]
    result = []
    for no, line in enumerate(lines, from_line + 1):
        result.append(
            f'{no:0{line_no_width}} {line}'
        )
        if no == line_no:
            result.append(
                f'{"-"*(line_no_width+column+1)}^'
            )

    return '\n'.join(result)



_conversion_powers = {
    'KB': 1,
    'MB': 2,
    'GB': 3,
    'TB': 4,
}


def cgroup_memory_usage(unit='B', cgroup_mem_file='/sys/fs/cgroup/memory/memory.usage_in_bytes'):
    """
    Returns the memory usage of the cgroup the Python interpreter is running in.

    This is handy for getting the memory usage inside a Docker container since they are
    implemented using cgroups.
    With conventional tools (e.g. psutil) this is not possible, because they often rely on
    stats reported by /proc, but that one reports metrics from the host system.
    """
    with open(cgroup_mem_file, 'r') as infile:
        usage_bytes = infile.readline()
    usage_bytes = int(usage_bytes)

    if usage_bytes == 0 or unit == 'B':
        return usage_bytes

    return usage_bytes / 1024 ** _conversion_powers[unit]



class MockTimeMonotonicGenerator:
    """
    Helper to mock `time.monotonic()` in tests.

    e.g.:

    class FooBar(TestCase):
        @mock.patch.object(time, 'monotonic', MockTimeMonotonicGenerator())
        def test_something(self):
            ...
    """

    def __init__(self, offset=1):
        self.offset = offset
        self.num = 0

    def __call__(self):
        self.num += self.offset
        return self.num



import json
import pprint


def pformat(value):
    """
    Format given object: Try JSON fist and fallback to pformat()
    (JSON dumps are nicer than pprint.pformat() ;)
    """
    try:
        value = json.dumps(value, indent=4, sort_keys=True, ensure_ascii=False)
    except TypeError:
        # Fallback if values are not serializable with JSON:
        value = pprint.pformat(value, width=120)

    return value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['bx_py_utils',
 'bx_py_utils.aws',
 'bx_py_utils.humanize',
 'bx_py_utils.test_utils']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['publish = bx_py_utils_tests.publish:publish']}

setup_kwargs = {
    'name': 'bx-py-utils',
    'version': '50',
    'description': 'Various Python utility functions',
    'long_description': '# Boxine - bx_py_utils\n\nVarious Python utility functions\n\n\n## Quickstart\n\n```bash\npip install bx_py_utils\n```\n\n\n## Existing stuff\n\nHere only a simple list about existing utilities.\nPlease take a look into the sources and tests for deeper informations.\n\n\n[comment]: <> (✂✂✂ auto generated start ✂✂✂)\n\n### bx_py_utils.anonymize\n\n* [`anonymize()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/anonymize.py#L15-L41) - Anonymize the given string with special handling for eMail addresses.\n\n### bx_py_utils.auto_doc\n\n* [`assert_readme()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/auto_doc.py#L121-L163) - Check and update README file with generate_modules_doc()\n* [`generate_modules_doc()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/auto_doc.py#L43-L118) - Generate a list of function/class information via pdoc.\n* [`get_code_location()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/auto_doc.py#L35-L40) - Return start and end line number for an object via inspect.\n\n#### bx_py_utils.aws.client_side_cert_manager\n\n* [`ClientSideCertManager()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/aws/client_side_cert_manager.py#L6-L72) - Helper to manage client-side TLS certificate via AWS Secrets Manager by\n\n#### bx_py_utils.aws.secret_manager\n\n* [`SecretsManager()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/aws/secret_manager.py#L4-L34) - Access AWS Secrets Manager values\n\n### bx_py_utils.compat\n\n* [`removeprefix()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/compat.py#L1-L7) - Backport of `removeprefix` from PEP-616 (Python 3.9+)\n* [`removesuffix()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/compat.py#L10-L16) - Backport of `removesuffix` from PEP-616 (Python 3.9+)\n\n### bx_py_utils.dict_utils\n\n* [`dict_get()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/dict_utils.py#L4-L25) - nested dict `get()`\n* [`pluck()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/dict_utils.py#L28-L40) - Extract values from a dict, if they are present\n\n### bx_py_utils.environ\n\n* [`cgroup_memory_usage()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/environ.py#L9-L25) - Returns the memory usage of the cgroup the Python interpreter is running in.\n\n### bx_py_utils.error_handling\n\n* [`print_exc_plus()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/error_handling.py#L14-L72) - Print traceback information with a listing of all the local variables in each frame.\n\n### bx_py_utils.graphql_introspection\n\n* [`introspection_query()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/graphql_introspection.py#L5-L25) - Generate GraphQL introspection query with variable nested depth.\n\n### bx_py_utils.hash_utils\n\n* [`url_safe_encode()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/hash_utils.py#L13-L22) - Encode bytes into a URL safe string.\n* [`url_safe_hash()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/hash_utils.py#L25-L46) - Generate a URL safe hash with `max_size` from given string/bytes.\n\n### bx_py_utils.html_utils\n\n* [`InvalidHtml()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/html_utils.py#L18-L41) - XMLSyntaxError with better error messages: used in validate_html()\n* [`pretty_format_html()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/html_utils.py#L69-L83) - Pretty format given HTML document via BeautifulSoup (Needs \'beautifulsoup4\' package)\n* [`validate_html()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/html_utils.py#L44-L66) - Validate a HTML document via XMLParser (Needs \'lxml\' package)\n\n#### bx_py_utils.humanize.pformat\n\n* [`pformat()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/humanize/pformat.py#L5-L16) - Format given object: Try JSON fist and fallback to pformat()\n\n#### bx_py_utils.humanize.time\n\n* [`human_timedelta()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/humanize/time.py#L14-L52) - Converts a time duration into a friendly text representation.\n\n### bx_py_utils.iteration\n\n* [`chunk_iterable()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/iteration.py#L4-L21) - Returns a generator that yields slices of iterable of the given `chunk_size`.\n\n### bx_py_utils.path\n\n* [`assert_is_dir()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/path.py#L5-L13) - Check if given path is a directory\n* [`assert_is_file()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/path.py#L16-L26) - Check if given path is a file\n\n### bx_py_utils.processify\n\n* [`processify()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/processify.py#L12-L53) - Decorator to run a function as a process.\n\n### bx_py_utils.stack_info\n\n* [`FrameNotFound()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/stack_info.py#L8-L13) - Base class for lookup errors.\n* [`last_frame_outside_path()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/stack_info.py#L16-L44) - Returns the stack frame that is the direct successor of given "file_path".\n\n#### bx_py_utils.test_utils.assertion\n\n* [`assert_equal()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/assertion.py#L52-L65) - Check if the two objects are the same. Display a nice diff, using `pformat()`\n* [`assert_text_equal()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/assertion.py#L68-L83) - Check if the two text strings are the same. Display a error message with a diff.\n* [`pformat_ndiff()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/assertion.py#L14-L24) - Generate a `ndiff` from two objects, using `pformat()`\n* [`pformat_unified_diff()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/assertion.py#L45-L49) - Generate a unified diff from two objects, using `pformat()`\n* [`text_ndiff()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/assertion.py#L6-L11) - Generate a `ndiff` between two text strings.\n* [`text_unified_diff()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/assertion.py#L38-L42) - Generate a unified diff between two text strings.\n\n#### bx_py_utils.test_utils.datetime\n\n* [`parse_dt()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/datetime.py#L4-L15) - Helper for easy generate a `datetime` instance via string.\n\n#### bx_py_utils.test_utils.filesystem_utils\n\n* [`FileWatcher()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/filesystem_utils.py#L6-L49) - Helper to record which new files have been created.\n\n#### bx_py_utils.test_utils.log_utils\n\n* [`RaiseLogUsage()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/log_utils.py#L4-L11) - A log handler, that raise an error on every log output.\n\n#### bx_py_utils.test_utils.mock_aws_secret_manager\n\n* [`SecretsManagerMock()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/mock_aws_secret_manager.py#L1-L16) - Mock for `bx_py_utils.aws.secret_manager.SecretsManager()`\n\n#### bx_py_utils.test_utils.mock_boto3session\n\n* [`MockedBoto3Session()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/mock_boto3session.py#L4-L46) - Mock for `boto3.session.Session()`\n\n#### bx_py_utils.test_utils.mocks3\n\nA simple mock for Boto3\'s S3 modules.\n\n* [`PseudoS3Client()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/mocks3.py#L39-L156) - Simulates a boto3 S3 client object in tests\n\n#### bx_py_utils.test_utils.requests_mock_assertion\n\n* [`assert_json_requests_mock()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/requests_mock_assertion.py#L30-L45) - Check the requests mock history. In this case all requests must be JSON.\n* [`assert_json_requests_mock_snapshot()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/requests_mock_assertion.py#L48-L55) - Check requests mock history via snapshot. Accepts only JSON requests.\n* [`assert_requests_mock()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/requests_mock_assertion.py#L58-L78) - Check the requests mock history. Accept mixed "text" and "JSON".\n* [`assert_requests_mock_snapshot()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/requests_mock_assertion.py#L81-L86) - Check requests mock history via snapshot. Accept mixed "text" and "JSON".\n\n#### bx_py_utils.test_utils.snapshot\n\nAssert complex output via auto updated snapshot files with nice diff error messages.\n\n* [`assert_html_snapshot()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/snapshot.py#L223-L261) - Assert "html" string via snapshot file with validate and pretty format\n* [`assert_py_snapshot()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/snapshot.py#L186-L220) - Assert complex python objects vio PrettyPrinter() snapshot file.\n* [`assert_snapshot()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/snapshot.py#L151-L183) - Assert given data serialized to JSON snapshot file.\n* [`assert_text_snapshot()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/snapshot.py#L115-L148) - Assert "text" string via snapshot file\n\n#### bx_py_utils.test_utils.time\n\n* [`MockTimeMonotonicGenerator()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/test_utils/time.py#L1-L19) - Helper to mock `time.monotonic()` in tests.\n\n### bx_py_utils.text_tools\n\n* [`cutout()`](https://github.com/boxine/bx_py_utils/blob/master/bx_py_utils/text_tools.py#L1-L36) - Mark a point in a long text by line no + column with context lines around.\n\n[comment]: <> (✂✂✂ auto generated end ✂✂✂)\n\n\n## Backwards-incompatible changes\n\n### v36 -> v37 - Outsourcing Django stuff\n\nWe split `bx_py_utils` and moved all Django related utilities into the separated project:\n\n* https://github.com/boxine/bx_django_utils\n\nSo, `bx_py_utils` is better usable in non-Django projects, because Django will not installed as decency of "bx_py_utils"\n\n\n## developing\n\nTo start developing e.g.:\n\n```bash\n~$ git clone https://github.com/boxine/bx_py_utils.git\n~$ cd bx_py_utils\n~/bx_py_utils$ make\nhelp                 List all commands\ninstall-poetry       install or update poetry\ninstall              install via poetry\nupdate               Update the dependencies as according to the pyproject.toml file\nlint                 Run code formatters and linter\nfix-code-style       Fix code formatting\ntox-listenvs         List all tox test environments\ntox                  Run pytest via tox with all environments\ntox-py36             Run pytest via tox with *python v3.6*\ntox-py37             Run pytest via tox with *python v3.7*\ntox-py38             Run pytest via tox with *python v3.8*\ntox-py39             Run pytest via tox with *python v3.9*\npytest               Run pytest\npytest-ci            Run pytest with CI settings\npublish              Release new version to PyPi\nclean                Remove created files from the test project\n```\n\n\n## License\n\n[MIT](LICENSE). Patches welcome!\n\n## About us\n\nWe’ve been rethinking the listening experience for kids and have created an ecosystem where haptic and listening experience are combined via smart technology - the Toniebox.\n\nWe are constantly looking for engineers to join our team in different areas. If you’d be interested in contributing to our platform, have a look at: https://tonies.com/jobs/\n\n## Links\n\n* https://pypi.org/project/bx-py-utils/\n',
    'author': 'Jens Diemer',
    'author_email': 'jens.diemer@boxine.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/boxine/bx_py_utils/',
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0.0',
}


setup(**setup_kwargs)



import logging


class RaiseLogUsage(logging.Handler):
    """
    A log handler, that raise an error on every log output.
    This is useful to check if tests generates not captured log output.
    """

    def handle(self, record):
        raise AssertionError(f'Missing log capture for: "{self.format(record)}"')



import re
import string


_LOWERCASES = string.ascii_lowercase + 'äöüß'
_UPPERCASES = string.ascii_uppercase + 'ÄÖÜẞ'
_OTHER = ' ' + string.punctuation
_ANONYMIZATION_TRANS = str.maketrans(
    _LOWERCASES + _UPPERCASES + string.digits + _OTHER,
    'x' * len(_LOWERCASES) + 'X' * len(_UPPERCASES) + '#' * len(string.digits) + '_' * len(_OTHER)
)
_RE_OTHER = re.compile(r'[^xX_\.\s/-]')


def anonymize(value: str) -> str:
    """
    Anonymize the given string with special handling for eMail addresses.

    >>> anonymize('Foo Bar')
    'Fxx_Xxr'
    >>> anonymize('This is a Test 123 Foo Bar #+"-! End')
    'Txxx_xx_x_Xxxx_###_Xxx_Xxx_______Xxd'
    >>> anonymize('a.mail-address@test.tld')
    'a_xxxx_xxxxxxs@test.tld'
    """
    assert isinstance(value, str)

    if '@' in value:
        value, at, domain = value.partition('@')
        if len(value) < 2:
            return value + at + domain

        return (
            value[:1] +
            _RE_OTHER.sub('@', value[1:-1].translate(_ANONYMIZATION_TRANS)) +
            value[-1:] +
            at + domain
        )

    value = f'{value[:1]}{value[1:-1].translate(_ANONYMIZATION_TRANS)}{value[-1:]}'
    return value
